/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   iftab.c
 *    Implementation of the FreeM transactions-in-flight symbol table
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2020 Coherent Logic Development LLC
 *
 *   Last modified: 18 March 2020
 * 
 **/

#include <stdlib.h>
#include "iftab.h"
#include "mpsdef.h"
#include "transact.h"

iftab *iftab_head;

iftab *iftab_insert(char *key, char *data, int tlevel)
{
    iftab *t;

    for (t = iftab_head; t != NULL; t = t->next) {
        
        if (stcmp (t->key, key) == 0) {

            /* the key already exists. this is now an update. */
            free (t->data);
            t->data = malloc (stlen (data) + 1);

            stcpy (t->data, data);
            t->killed = FALSE;

            return t;
        }

    }

    /* key did not exist. this is now a true insert. */

    t = (iftab *) malloc (sizeof (iftab));
    t->key = malloc (stlen (key) + 1);
    t->data = malloc (stlen (data) + 1);

    stcpy (t->key, key);
    stcpy (t->data, data);

    t->killed = FALSE;
    t->tlevel = tlevel;

    t->next = iftab_head;
    iftab_head = t;

    return t;
}

iftab *iftab_retrieve(char *key, char *data)
{
    iftab *t;

    for (t = iftab_head; t != NULL; t = t->next) {
        if (stcmp (t->key, key) == 0) {
            data = (char *) malloc (stlen (t->data) + 1);
            stcpy (data, t->data);

            return t;
        }
    }

    return (iftab *) NULL;
}

void iftab_delete(char *key)
{
    iftab *t = iftab_head;
    iftab *p;


    if ((t != (iftab *) NULL) && (stcmp (t->key, key) == 0)) {
        iftab_head = t->next;

        free (t->key);
        free (t->data);
        free (t);
        return;
    }

    while ((t != NULL) && (stcmp (t->key, key) != 0)) {
        p = t;
        t = t->next;
    }

    if (t == NULL) return;

    p->next = t->next;
    free (t->key);
    free (t->data);
    free (t);

    return;
}

iftab *iftab_kill(char *key)
{
    iftab *t;

    for(t = iftab_head; t != NULL; t = t->next) {
        if (stcmp (t->key, key) == 0) {
            t->killed = TRUE;

            return t;
        }
    }

    t = iftab_insert (key, "\201", tp_level);
    t->killed = TRUE;

    return t;
}

void iftab_pop_tlevel(int tlevel)
{
    iftab *t;

    for (t = iftab_head; t != NULL; t = t->next) {
        if (t->tlevel == tlevel) iftab_delete (t->key);
    }
}